import numpy as np
import math

def updateBackgroundAtInterval(i, hyperparameters, background, wellPositions, wellNumber, initialCurFrame, firstFrame, trackingHeadTailAllAnimals, frame):
  
  if i % hyperparameters["updateBackgroundAtInterval"] == 0:
    showImages = False
    firstFrameToShow = 10
    if showImages and i > firstFrameToShow:
      import zebrazoom.code.util as util
      util.showFrame(background, title='background before')
    xvalues = [trackingHeadTailAllAnimals[0, i-firstFrame][k][0] for k in range(0, len(trackingHeadTailAllAnimals[0, i-firstFrame]))]
    yvalues = [trackingHeadTailAllAnimals[0, i-firstFrame][k][1] for k in range(0, len(trackingHeadTailAllAnimals[0, i-firstFrame]))]
    xmin = min(xvalues)
    xmax = max(xvalues)
    ymin = min(yvalues)
    ymax = max(yvalues)
    dist = 1 * math.sqrt((xmax - xmin) ** 2 + (ymax - ymin) ** 2)
    xmin = int(xmin - dist) if xmin - dist >= 0 else 0
    xmax = int(xmax + dist) if xmax + dist < len(frame[0]) else len(frame[0]) - 1
    ymin = int(ymin - dist) if ymin - dist >= 0 else 0
    ymax = int(ymax + dist) if ymax + dist < len(frame) else len(frame) - 1
    if xmin != xmax and ymin != ymax:
      partOfBackgroundToSave = background[wellPositions[wellNumber]["topLeftY"]+ymin:wellPositions[wellNumber]["topLeftY"]+ymax, wellPositions[wellNumber]["topLeftX"]+xmin:wellPositions[wellNumber]["topLeftX"]+xmax].copy() # copy ???
      if showImages and i > firstFrameToShow:
        util.showFrame(partOfBackgroundToSave, title='partOfBackgroundToSave')
    background[wellPositions[wellNumber]["topLeftY"]:wellPositions[wellNumber]["topLeftY"]+wellPositions[wellNumber]["lengthY"], wellPositions[wellNumber]["topLeftX"]:wellPositions[wellNumber]["topLeftX"]+wellPositions[wellNumber]["lengthX"]] = initialCurFrame.copy()
    if showImages and i > firstFrameToShow:
      util.showFrame(background, title='background middle')
    if xmin != xmax and ymin != ymax:
      background[wellPositions[wellNumber]["topLeftY"]+ymin:wellPositions[wellNumber]["topLeftY"]+ymax, wellPositions[wellNumber]["topLeftX"]+xmin:wellPositions[wellNumber]["topLeftX"]+xmax] = partOfBackgroundToSave
    if showImages and i > firstFrameToShow:
      util.showFrame(background, title='background after')
  
  return background



import cv2
import numpy as np

def findCenterByIterativelyDilating(initialContour, lenX, lenY):
  x = 0
  y = 0
  
  xmin = lenX
  ymin = lenY
  xmax = 0
  ymax = 0
  for pt in initialContour:
    if pt[0][0] < xmin:
      xmin = pt[0][0]
    if pt[0][1] < ymin:
      ymin = pt[0][1]
    if pt[0][0] > xmax:
      xmax = pt[0][0]
    if pt[0][1] > ymax:
      ymax = pt[0][1]
  
  for pt in initialContour:
    pt[0][0] = pt[0][0] - xmin
    pt[0][1] = pt[0][1] - ymin
  
  image = np.zeros((ymax - ymin, xmax - xmin))
  image[:, :] = 255
  image = image.astype(np.uint8)
  kernel  = np.ones((3, 3), np.uint8)
  if type(initialContour) != int:
    cv2.fillPoly(image, pts =[initialContour], color=(0))
    image[:,0] = 255
    image[0,:] = 255
    image[:, len(image[0])-1] = 255
    image[len(image)-1, :]    = 255
    nbBlackPixels = 1
    dilateIter = 0
    while nbBlackPixels > 0:
      dilateIter   = dilateIter + 1
      dilatedImage = cv2.dilate(image, kernel, iterations=dilateIter)
      nbBlackPixels = cv2.countNonZero(255-dilatedImage)
    dilateIter   = dilateIter - 1
    dilatedImage = cv2.dilate(image, kernel, iterations=dilateIter)
    dilatedImage[:,0] = 255
    dilatedImage[0,:] = 255
    dilatedImage[:, len(dilatedImage[0])-1] = 255
    dilatedImage[len(dilatedImage)-1, :]    = 255
    contours, hierarchy = cv2.findContours(dilatedImage, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    maxContour = 0
    maxContourArea = 0
    for contour in contours:
      contourArea = cv2.contourArea(contour)
      if contourArea < int((xmax - xmin) * (ymax - ymin) * 0.8):
        if contourArea > maxContourArea:
          maxContourArea = contourArea
          maxContour     = contour
    M = cv2.moments(maxContour)
    if M['m00']:
      x = int(M['m10']/M['m00'])
      y = int(M['m01']/M['m00'])
  
  return [x + xmin, y + ymin]



import numpy as np
import cv2
from zebrazoom.code.trackingFolder.trackingFunctions import calculateAngle
from zebrazoom.code.trackingFolder.trackingFunctions import distBetweenThetas
from zebrazoom.code.trackingFolder.trackingFunctions import assignValueIfBetweenRange
import math
from scipy.interpolate import UnivariateSpline
from numpy import linspace
import os.path
import csv
from zebrazoom.code.getImage.headEmbededFrame import headEmbededFrame


def getAccentuateFrameForManualPointSelect(image, hyperparameters):
  if hyperparameters["accentuateFrameForManualTailExtremityFind"]:
    frame = image.copy()
    quartileChose = 0.01
    lowVal  = int(np.quantile(frame, quartileChose))
    highVal = int(np.quantile(frame, 1 - quartileChose))
    frame[frame < lowVal]  = lowVal
    frame[frame > highVal] = highVal
    frame = frame - lowVal
    mult  = np.max(frame)
    frame = frame * (255/mult)
    frame = frame.astype(int)
    frame = (frame / np.linalg.norm(frame))*255
    return frame
  else:
    return image


def findTailTipByUserInput(frame, frameNumber, videoPath, hyperparameters, wellNumber, wellPositions):
  from PyQt5.QtWidgets import QApplication

  import zebrazoom.code.util as util

  plus = 0

  def tailNotStraight(frameWidget):
    nonlocal plus
    plus += 1
    util.setPixmapFromCv(headEmbededFrame(videoPath, frameNumber + plus, wellNumber, wellPositions, hyperparameters)[0], frameWidget, zoomable=True)
  return list(util.getPoint(np.uint8(frame * 255), "Click on tail tip", zoomable=True, extraButtons=(("Tail is not straight", tailNotStraight, False),),
                            dialog=not hasattr(QApplication.instance(), 'window')))

def findHeadPositionByUserInput(frame, frameNumber, videoPath, hyperparameters, wellNumber, wellPositions):
  from PyQt5.QtWidgets import QApplication

  import zebrazoom.code.util as util

  plus = 0

  def tailNotStraight(frameWidget):
    nonlocal plus
    plus += 1
    util.setPixmapFromCv(headEmbededFrame(videoPath, frameNumber + plus, wellNumber, wellPositions, hyperparameters)[0], frameWidget, zoomable=True)
  return list(util.getPoint(np.uint8(frame * 255), "Click on the base of the tail", zoomable=True, extraButtons=(("Tail is not straight", tailNotStraight, False),),
                            dialog=not hasattr(QApplication.instance(), 'window')))


def getTailTipByFileSaved(hyperparameters,videoPath):
  ix = -1
  iy = -1
  with open(videoPath+'.csv') as csv_file:
    csv_reader = csv.reader(csv_file, delimiter=',')
    line_count = 0
    for row in csv_reader:
      if len(row):
        ix = row[0]
        iy = row[1]
  return [int(ix),int(iy)]

def getHeadPositionByFileSaved(videoPath):
  ix = -1
  iy = -1
  with open(videoPath+'HP.csv') as csv_file:
    csv_reader = csv.reader(csv_file, delimiter=',')
    line_count = 0
    for row in csv_reader:
      if len(row):
        ix = row[0]
        iy = row[1]
  return [int(ix),int(iy)]
  


import numpy as np
import cv2
from zebrazoom.code.trackingFolder.trackingFunctions import calculateAngle
from zebrazoom.code.trackingFolder.trackingFunctions import distBetweenThetas
from zebrazoom.code.trackingFolder.trackingFunctions import assignValueIfBetweenRange
import math
from scipy.interpolate import UnivariateSpline
from numpy import linspace
import os.path
import csv
import pdb


def smoothTail(points, nbTailPoints):
  
  y = points[0]
  x = linspace(0, 1, len(y))
  
  if len(x) > 3:    
  
    s = UnivariateSpline(x, y, s=10)
    xs = linspace(0, 1, nbTailPoints)
    newX = s(xs)

    y = points[1]
    x = linspace(0, 1, len(y))
    s = UnivariateSpline(x, y, s=10)
    xs = linspace(0, 1, nbTailPoints)
    newY = s(xs)
    
  else:
  
    newX = x
    newY = y
  
  return [newX, newY]

def appendPoint(x, y, points):
  curPoint = np.zeros((2, 1))
  curPoint[0] = x;
  curPoint[1] = y;
  points = np.append(points, curPoint, axis=1)
  return points

def findNextPoints(depth,x,y,frame,points,angle,maxDepth,hyperparameters,debug):
  
  step     = hyperparameters["centerOfMassParamStep"]
  segStep  = hyperparameters["centerOfMassParamSegStep"]
  halfDiam = hyperparameters["centerOfMassParamHalfDiam"]
  
  if debug:
    frameDisplay = frame.copy()
    frameDisplay = cv2.cvtColor(frameDisplay, cv2.COLOR_GRAY2RGB)
    
  lenX = len(frame[0]) - 1
  lenY = len(frame) - 1

  thetaDiffAccept = 0.4
  pixTotMax = 1000000
  maxTheta  = angle

  xNew = assignValueIfBetweenRange(int(x + step * (math.cos(angle))), 0, lenX)
  yNew = assignValueIfBetweenRange(int(y + step * (math.sin(angle))), 0, lenY)
  
  # if debug:
    # cv2.circle(frameDisplay, (xNew, yNew), 1, (0,0,0),   -1)

  framecopy = frame.copy()
  
  xmin = assignValueIfBetweenRange(xNew - halfDiam, 0, lenX-1)
  xmax = assignValueIfBetweenRange(xNew + halfDiam, 0, lenX-1)
  ymin = assignValueIfBetweenRange(yNew - halfDiam, 0, lenY-1)
  ymax = assignValueIfBetweenRange(yNew + halfDiam, 0, lenY-1)
  
  if debug:
    cv2.circle(frameDisplay, (xmin, ymin), 1, (255,0,0),   -1)
    cv2.circle(frameDisplay, (xmax, ymax), 1, (255,0,0),   -1)
    cv2.circle(frameDisplay, (x, y),       1, (0,255,0),   -1)
  
  if (xmin==0 or xmin==lenX-1 or xmax==0 or xmax==lenX-1 or ymin==0 or ymin==lenY-1 or ymax==0 or ymax==lenY-1):
    return (points,0)
  
  framecopy[0:ymin, :]              = 255
  framecopy[ymax:lenY+1, :]         = 255
  framecopy[:, 0:xmin]              = 255
  framecopy[:, xmax:lenX+1]         = 255
  
  if debug:
    import zebrazoom.code.util as util

    util.showFrame(framecopy, title='Frame2')
  
  (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(framecopy)
  
  xNew = headPosition[0]
  yNew = headPosition[1]
  
  theta = calculateAngle(x, y, xNew, yNew)
  
  xNew = assignValueIfBetweenRange(int(x + segStep  * (math.cos(theta)) ), 0, lenX)
  yNew = assignValueIfBetweenRange(int(y + segStep  * (math.sin(theta)) ), 0, lenY)
  
  # Calculates distance between new and old point
  distSubsquentPoints = math.sqrt((xNew - x)**2 + (yNew - y)**2)
  
  if debug:
    import zebrazoom.code.util as util

    cv2.circle(frameDisplay, (xNew, yNew), 1, (0,0,255),   -1)
    util.showFrame(frameDisplay, title='Frame2')
    
    util.showFrame(framecopy, title='Frame2')
    
  points = appendPoint(xNew, yNew, points)
  newTheta = calculateAngle(x,y,xNew,yNew)
  if distSubsquentPoints > 0 and depth < maxDepth:
    (points,nop) = findNextPoints(depth+distSubsquentPoints,xNew,yNew,frame,points,newTheta,maxDepth,hyperparameters,debug)
  
  return (points,newTheta)


def centerOfMassTailTracking(headPosition,nbTailPoints,i,thresh1,frame,hyperparameters,heading,maxDepth):
  
  x = headPosition[0]
  y = headPosition[1]
  
  initialAngle = hyperparameters["headEmbededParamInitialAngle"]
  
  gaussian_blur = hyperparameters["headEmbededParamGaussianBlur"]
  frame = cv2.GaussianBlur(frame, (gaussian_blur, gaussian_blur), 0)
  points = np.zeros((2, 0))
  
  # if i > 63033:
    # (points, lastFirstTheta2) = findNextPoints(0,x,y,frame,points,initialAngle,maxDepth,hyperparameters,True)
  # else:
  (points, lastFirstTheta2) = findNextPoints(0,x,y,frame,points,initialAngle,maxDepth,hyperparameters,False)
  
  output = np.zeros((1, nbTailPoints, 2))
  
  points = np.insert(points, 0, headPosition, axis=1)
  
  points = points[:,0:(len(points[0])-1)]
  
  points = smoothTail(points, nbTailPoints)

  for idx, x in enumerate(points[0]):
    output[0][idx][0] = x
    output[0][idx][1] = points[1][idx]

  return output


def centerOfMassTailTrackFindMaxDepth(headPosition,nbTailPoints,i,x,y,thresh1,frame,hyperparameters,oppHeading,tailTip):
  
  gaussian_blur = hyperparameters["headEmbededParamGaussianBlur"]
  frame = cv2.GaussianBlur(frame, (gaussian_blur, gaussian_blur), 0)
  angle = hyperparameters["headEmbededParamInitialAngle"]
  points = np.zeros((2, 0))
  
  (points, lastFirstTheta2) = findNextPoints(0,x,y,frame,points,angle,300,hyperparameters,False)
  
  distToTip        = np.full((200),10000)
  curTailLengthTab = np.full((200),10000)
  curTailLength  = 0
  k = 0
  
  distFromHeadToTip = abs(math.sqrt((x-tailTip[0])**2 + (y-tailTip[1])**2))
  while (curTailLength < 1.5*distFromHeadToTip) and (k < len(points[0])-1):
    curTailLength = curTailLength + abs(math.sqrt((points[0,k]-points[0,k+1])**2 + (points[1,k]-points[1,k+1])**2))
    distToTip[k]  = abs(math.sqrt((points[0,k]-tailTip[0])**2 + (points[1,k]-tailTip[1])**2))
    curTailLengthTab[k] = curTailLength
    k = k + 1
  
  minDistToTip    = 1000000
  indMinDistToTip = 0
  for idx, dist in enumerate(distToTip):
    if dist < minDistToTip:
      minDistToTip = dist
      indMinDistToTip = idx
  
  return (curTailLengthTab[indMinDistToTip] )



import cv2
import math
import numpy as np

from zebrazoom.code.trackingFolder.trackingFunctions import calculateAngle
from zebrazoom.code.trackingFolder.trackingFunctions import distBetweenThetas

def computeHeading(thresh1, x, y, headSize, hyperparameters):
  
  videoWidth  = hyperparameters["videoWidth"]
  videoHeight = hyperparameters["videoHeight"]
  
  x = x + int(headSize)
  y = y + int(headSize)
  
  paddedImage = np.zeros((len(thresh1) + 2 * int(headSize), len(thresh1[0]) + 2 * int(headSize)))
  paddedImage[:, :] = 255
  paddedImage[int(headSize):len(thresh1)+int(headSize), int(headSize):len(thresh1[0])+int(headSize)] = thresh1
  thresh1 = paddedImage
  
  thresh1 = thresh1.astype(np.uint8)

  ymin  = y - headSize
  ymax  = y + headSize
  xmin  = x - headSize
  xmax  = x + headSize
  
  img = thresh1[int(ymin):int(ymax), int(xmin):int(xmax)]
  
  img[0,:] = 255
  img[len(img)-1,:] = 255
  img[:,0] = 255
  img[:,len(img[0])-1] = 255
  
  y, x = np.nonzero(img)
  x = x - np.mean(x)
  y = y - np.mean(y)
  coords = np.vstack([x, y])
  cov = np.cov(coords)
  evals, evecs = np.linalg.eig(cov)
  sort_indices = np.argsort(evals)[::-1]
  x_v1, y_v1 = evecs[:, sort_indices[0]]  # Eigenvector with largest eigenvalue
  x_v2, y_v2 = evecs[:, sort_indices[1]]
  scale = 20
  theta = calculateAngle(0, 0, x_v1, y_v1)
  theta = (theta - math.pi/2) % (2 * math.pi)
  
  if hyperparameters["debugHeadingCalculation"]:
    import zebrazoom.code.util as util

    img2 = img.copy()
    img2 = cv2.cvtColor(img2, cv2.COLOR_GRAY2BGR)
    cv2.line(img2, (int(len(img2[0])/2), int(len(img2)/2)), (int(len(img[0])/2 + 20 * math.cos(theta)), int(len(img)/2 + 20 * math.sin(theta))), (255,0,0), 1)
    util.showFrame(img2, title='imgForHeadingCalculation')
  
  return theta


def calculateHeading(x, y, i, thresh1, thresh2, takeTheHeadClosestToTheCenter, hyperparameters, previousFrameHeading=0):

  nbList = hyperparameters["nbList"]
  expDecreaseFactor = hyperparameters["expDecreaseFactor"]
  firstFrame = hyperparameters["firstFrame"]
  lastFrame = hyperparameters["lastFrame"]
  headSize = hyperparameters["headSize"]
  debugTracking = hyperparameters["debugTracking"]
  headEmbeded = hyperparameters["headEmbeded"]
  
  if hyperparameters["debugHeadingCalculation"]:
    import zebrazoom.code.util as util

    util.showFrame(thresh1, title='thresh1')
    util.showFrame(thresh2, title='thresh2')
  
  cx = 0
  cy = 0
  cxNew = 0
  cyNew = 0
    
  thresh1[0,:]                 = 255
  thresh1[len(thresh1)-1,:]    = 255
  thresh1[:,0]                 = 255
  thresh1[:,len(thresh1[0])-1] = 255
  
  contours, hierarchy = cv2.findContours(thresh1,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
  for contour in contours:
    dist = cv2.pointPolygonTest(contour, (x, y), True)
    if dist >= 0:
      M = cv2.moments(contour)
      if M['m00']:
        cx = int(M['m10']/M['m00'])
        cy = int(M['m01']/M['m00'])
        bodyContour = contour
      else:
        cx = 0
        cy = 0
      if hyperparameters["debugHeadingCalculation"]:
        print("area:", area)
        print("dist, cx, cy:", dist, cx, cy)
        print("x, y:", x, y)
  
  if hyperparameters["debugHeadingCalculation"]:
    print("x, y, cx, cy:", x, y, cx, cy)
  
  heading = computeHeading(thresh2, x, y, headSize, hyperparameters)
  if hyperparameters["debugHeadingCalculation"]:
    print("previousFrameHeading:", previousFrameHeading)
    print("math.sqrt((x - cx)**2 + (y - cy)**2):", math.sqrt((x - cx)**2 + (y - cy)**2))
  if math.sqrt((x - cx)**2 + (y - cy)**2) > 1:
    lastFirstTheta = calculateAngle(x, y, cx, cy)
  else:
    lastFirstTheta = (previousFrameHeading + math.pi) % (2 * math.pi)
  
  headingApproximate = lastFirstTheta
  headingPreciseOpt1 = heading
  headingPreciseOpt2 = (heading + math.pi) % (2*math.pi)
  diffAngle1 = distBetweenThetas(headingApproximate, headingPreciseOpt1)
  diffAngle2 = distBetweenThetas(headingApproximate, headingPreciseOpt2)
  if (diffAngle1 > diffAngle2):
    heading = headingPreciseOpt1
  else:
    heading = headingPreciseOpt2
  
  if hyperparameters["debugHeadingCalculation"]:
    print("headingApproximate:", headingApproximate)
    print("headingPreciseOpt1:", headingPreciseOpt1)
    print("headingPreciseOpt2:", headingPreciseOpt2)
    print("diffAngle1:", diffAngle1)
    print("diffAngle2:", diffAngle2)
    print("heading:", heading)
  
  return [heading, lastFirstTheta]


def calculateHeadingSimple(x, y, thresh2, hyperparameters):
  
  headSize = hyperparameters["headSize"]
  heading = computeHeading(thresh2, x, y, headSize, hyperparameters)
  
  return heading
